"""A creation workflow that adds an existing switch to the service database."""

from orchestrator import step, workflow
from orchestrator.forms import FormPage
from orchestrator.targets import Target
from orchestrator.types import FormGenerator, State, SubscriptionLifecycle, UUIDstr
from orchestrator.workflow import StepList, begin, done
from orchestrator.workflows.steps import resync, set_status, store_process_subscription

from gso.products import ProductName
from gso.products.product_blocks.switch import SwitchModel
from gso.products.product_types.site import Site
from gso.products.product_types.switch import ImportedSwitchInactive
from gso.services.partners import get_partner_by_name
from gso.services.subscriptions import get_product_id_by_name
from gso.utils.shared_enums import Vendor
from gso.utils.types.ip_address import PortNumber


def _initial_input_form_generator() -> FormGenerator:
    class ImportSwitch(FormPage):
        fqdn: str
        ts_port: PortNumber
        site: UUIDstr
        switch_vendor: Vendor
        switch_model: SwitchModel

    user_input = yield ImportSwitch
    return user_input.model_dump() | {"partner": "GEANT"}


@step("Create subscription")
def create_subscription(partner: str) -> State:
    """Create a new subscription object."""
    partner_id = get_partner_by_name(partner)["partner_id"]
    product_id = get_product_id_by_name(ProductName.IMPORTED_SWITCH)
    subscription = ImportedSwitchInactive.from_product_id(product_id, partner_id)

    return {"subscription": subscription, "subscription_id": subscription.subscription_id}


@step("Initialize subscription")
def initialize_subscription(
    subscription: ImportedSwitchInactive,
    fqdn: str,
    ts_port: PortNumber,
    site: UUIDstr,
    switch_vendor: Vendor,
    switch_model: SwitchModel,
) -> State:
    """Initialize the switch subscription using input data."""
    subscription.switch.fqdn = fqdn
    subscription.switch.ts_port = ts_port
    subscription.switch.site = Site.from_subscription(site).site
    subscription.switch.switch_vendor = switch_vendor
    subscription.switch.switch_model = switch_model

    return {"subscription": subscription}


@workflow("Create Imported Switch", initial_input_form=_initial_input_form_generator, target=Target.CREATE)
def create_imported_switch() -> StepList:
    """Create an imported switch without provisioning it."""
    return (
        begin
        >> create_subscription
        >> store_process_subscription(Target.CREATE)
        >> initialize_subscription
        >> set_status(SubscriptionLifecycle.ACTIVE)
        >> resync
        >> done
    )
